/* A client for connecting to a remote Eden server */

#include "eden_glue.h"

/* Function interfaces are documented in eden_glue.h */

/*** Functions for Connecting and Disconnecting *************************/
struct sockaddr_in eden_sockaddr_in(char *address, int port) {
    struct sockaddr_in eden_addr;
    eden_addr.sin_family = AF_INET;
    /* If a port number was given, we use that.  Otherwise, use the
     * default defined at the top of the file. */
    eden_addr.sin_port = htons(port);
    inet_aton(address, &(eden_addr.sin_addr));
    memset(eden_addr.sin_zero, '\0', sizeof eden_addr.sin_zero);
    return eden_addr;
}


SOCKFD connect_to_eden(struct sockaddr_in sockaddr) {
    SOCKFD eden_sock = socket(PF_INET, TYPE_OF_PROTOCOL, 0);
    int trap = connect(
            eden_sock, (struct sockaddr *)&sockaddr, sizeof sockaddr);
    return eden_sock;
}

void disconnect_from_eden(SOCKFD eden_sock) { close(eden_sock); };


/*** Functions for Constructing Commands *********************************/
size_t eden_cmd_len(const char *cmdname) {
    return strlen(CMDPRE) + strlen(cmdname) + strlen(CMDPOST);
}

void eden_cmd_wrap(char* result, const char *cmdname) {
    /* wrap a command in the standard prefix and suffix */
    strcat(result, CMDPRE);
    strcat(result, cmdname);
    strcat(result, CMDPOST);
}

size_t eden_cmd_notation_len(const char *notation) {
    /* How long will the notation line be? */
    return 1 + strlen(notation) + 2; /* "%" notation "\r\n" */
}

void eden_cmd_set_notation(char* result, const char *notation) {
    /* Add notation line to the result */
    strcat(result, "%");
    strcat(result, notation);
    strcat(result, "\r\n");
}

/*** Functions for communicating with dtkeden ***************************/
int eden_login(SOCKFD eden_sock, const char *hostname) {
    /* -> 0 for success */
    char login_command[1000] = "";   // 999 characters should be plenty
    eden_cmd_set_notation(login_command, "eden");
    eden_cmd_wrap(login_command, CMD_LOGIN);
    strcat(login_command, hostname);
    size_t sendlen = strlen(login_command);
    int sent = send(eden_sock, login_command, sendlen, 0);
    int sent2 = send(eden_sock, "\r\n", 2, 0);
    return 0;
}


int eden_run( SOCKFD eden_sock, const char *hostname,
              const char *notation, const char *cmd)
{
    /* -> 0 for success */

    /* Find out how long our message is going to be */
    size_t notation_len = eden_cmd_notation_len(notation);
    size_t cmd_len = strlen(cmd) + 2;       /* We'll be adding \r\n */
    size_t dtkcmd_len = eden_cmd_len(CMD_EOF);
    size_t hostname_len = strlen(hostname);
    size_t message_len = notation_len + cmd_len + dtkcmd_len + hostname_len;

    /* Create empty message and populate it */
    char message[message_len + 1];
    message[0] = '\0'; /* Lest I segfault */
    eden_cmd_set_notation(message, notation);   /* Start with notation */
    strcat(message, cmd);        /* Then add the command we want to send */
    strcat(message, "\r\n");
    eden_cmd_wrap(message, CMD_EOF); /* Then add the EOF command */
    strcat(message, hostname);  /* And finally append the local hostname */

    /* Send the message */
    int sent1 = send(eden_sock, message, strlen(message), 0);

    /* Now send a \r\n */
    int sent2 = send(eden_sock, "\r\n", 2, 0);
    return 0;
}


int eden_set_int(SOCKFD eden_sock, const char *hostname,
                 const char *obsname, int value)
{
    /* Work out how long the command will be */
    char value_str[17];
    sprintf(value_str, "%d", value);
    size_t value_len = strlen(value_str);
    size_t obsname_len = strlen(obsname);
    size_t total_len = obsname_len + 1 + value_len + 1;

    /* Initialise the command */
    char cmd[total_len];
    cmd[0] = '\0';

    /* cmd = obsname + "=" + value_str + ";" */
    strcat(cmd, obsname);
    strcat(cmd, "=");
    strcat(cmd, value_str);
    strcat(cmd, ";");

    /* Run it */
    return eden_run(eden_sock, hostname, "eden", cmd);
}
