/* Interface for a client for a running remote dtkeden server.
 * Author: Nick Booker
 *
 * Currently, this client can only save data in Eden.
 * Ideally, it would also fetch data from Eden, but that's a future
 * extension!
 *
 * This will probably break badly in all but the simplest of cases,
 * but it should be enough to demonstrate what I'm doing.
 */


/*** INCLUDES ************************************************************/
/* These are the dependencies of the library eden_glue.c */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>


/*** USEFUL MACROS *******************************************************/

#define EDEN_PORT 9000
/* EDEN_PORT: The port we will use by default to connect to dtkeden */

#define SOCKFD int
/* SOCKFD: An alias to int, just so we know when we're referring to or
 *         expecting a file descriptor for a socket
 */

#define TYPE_OF_PROTOCOL SOCK_STREAM
/* TYPE_OF_PROTOCOL: If we suddenly decide to use a datagram protocol,
 *                   update this.
 */


/*** Macros for Building Command Strings *********************************/
#define CMDPRE "@#$%"    /* The prefix for all dtkeden commands */
#define CMDPOST "%$#@"   /* The suffix for all dtkeden commands */

#define CMD_LOGIN "EOU"  /* Sent (among other things) to login */
#define CMD_EOF "EOF"    /* Terminates a normal Eden command   */
#define CMD_ECS "ECS"    /* Something to do with synchronised mode */
#define CMD_EOS "EOS"    /* Sorry, couldn't understand this one.   */

/*************************************************************************/
/* THE FUNCTION INTERFACES                                               */

/*** Functions for Connecting to Dtkeden Servers ***/

struct sockaddr_in eden_sockaddr_in(char *address, int port);
/*  An simpler interface for creating a sockaddr_in structure.
 */

SOCKFD connect_to_eden(struct sockaddr_in sockaddr);
/*  Connect the given socket address struct, bind to the Eden port and
 *  return a file descriptor.
 */

void disconnect_from_eden(SOCKFD eden_sock);
/*  Terminate communications with Eden
 */

/***   ***   ***   ***   ***   ***   ***   ***   ***   ***   ***   ***   */
/*** Functions for Constructing Dtkeden Commands ***/

size_t eden_cmd_len(const char *cmdname);
/*  Get what will be the length of the command built from cmdname
 */


void eden_cmd_wrap(char* result, const char *cmdname);
/*  Set result to be cmdname wrapped in the dtkeden command delimiters.
 *  *result: A pointer to a char[eden_cmd_len(cmdname)+1] = '\0'.
 */


/***   ***   ***   ***   ***   ***   ***   ***   ***   ***   ***   ***   */
/*** Generic dtkeden Commands ***/

int eden_login(SOCKFD eden_sock, const char *hostname);
/*  => trap code.  eden_trap is also set to same value.
 *  eden_sock: The socket connected to the dtkeden server
 *  hostname:  The name of your host, for identifying yourself to dtkeden
 */


int eden_run(
        SOCKFD      eden_sock,
        const char *hostname,
        const char *notation,
        const char *cmd
        );
/*  => trap code.  eden_trap is also set to same value.
 *  eden_sock: The socket connected to the dtkeden server
 *  hostname:  The name of your host, for identifying yourself to dtkeden
 *  notation:  The name of the notation (probably "eden")
 *  cmd:       The command (e.g "a=1000;") that you want to send to eden
 */


int eden_set_int(
              SOCKFD  eden_sock,
        const char   *hostname,
        const char   *obsname,
              int     value);
/* => trap code.  eden_trap is also set to same value.
 *  A slightly nicer way of setting an integer observable than calling
 *  eden_run directly.
 *
 *  eden_sock: The socket connected to the dtkeden server
 *  hostname:  The name of your host, for identifying yourself to dtkeden
 *  obsname:   The name of the Eden observable you want to set.
 *  value:     The integer value you want to set in Eden.
 *
 *  This is destructive -- any previous definition of obsname is erased.
 */

