/* Code comes and goes in here
 * It's just for checking the various bits of eden_glue work.
 * A lot of things are hard-coded and specific to the network or machine
 * on which they were tested!
 */

#include "eden_glue.h"

void test_eden_sockaddr_in() {
    printf("eden_sockaddr_in(\"127.0.0.1\", EDEN_PORT) ~> sai1 ... \n");
    struct sockaddr_in sai1 = eden_sockaddr_in("127.0.0.1", EDEN_PORT);
    printf("  inet_ntoa(sai1.sin_addr) == ");
    char *addr = inet_ntoa(sai1.sin_addr);
    printf("\"%s\"\n", addr);
    printf("  ntohs(sai1.sin_port) == ");
    uint16_t port = ntohs(sai1.sin_port);
    printf("%d\n", port);
}

void test_server_conn() {
    printf("Get http://localhost:8000/hello.txt\n");
    struct sockaddr_in mybox_addr = eden_sockaddr_in("127.0.0.1", 8000);
    SOCKFD mybox = connect_to_eden(mybox_addr);
    char *msg = "GET /hello.txt HTTP/1.1\nHost: 127.0.0.1\n\n";
    int len = strlen(msg);
    printf("The size of the request is supposedly %d.\n", len);
    int sent_bytes = send(mybox, msg, len, 0);
    printf("Sent %d\n", sent_bytes);
    char recvbuff[100];
    printf("Boo\n");
    int received = recv(mybox, recvbuff, 99, 0);
    recvbuff[received] = '\0';
    printf(recvbuff);
    close(mybox);
}

void test_eden_cmd_wrap() {
    printf("Test dtkeden_cmd_wrap(CMD_LOGIN) => @#$%%EOU%$#@\n");
    int wrapped_len = eden_cmd_len(CMD_LOGIN);
    printf("  wrapped_len == %d\n", wrapped_len);
    char result[wrapped_len + 1];
    result[0] = '\0';
    eden_cmd_wrap(result, CMD_LOGIN);
    printf("  result == %s\n", result);
}


void test_eden_login() {
    /* Checking whether this actually works involves interrogating the
     * server side manually; and possibly watching the traffic with
     * a sniffer tool */
    printf("Test things don't go pear shaped when we try to log in.\n");
    printf(" Connecting to merlin-wired...\n");
    struct sockaddr_in eden_addr = eden_sockaddr_in("192.168.1.6", EDEN_PORT);
    SOCKFD eden = connect_to_eden(eden_addr);
    printf(" Logging in...\n");
    int succ = eden_login(eden, "deepthought");
    close(eden);
}


void test_eden_run() {
    /* Checking whether this actually works involves interrogating the
     * server side manually; and possibly watching the traffic with
     * a sniffer tool */
    printf("Make sure we can send a command to eden after logging in.\n");
    printf(" Connecting...\n");
    struct sockaddr_in ea = eden_sockaddr_in("192.168.1.6", EDEN_PORT);
    SOCKFD eden = connect_to_eden(ea);
    printf(" Logging in...\n");
    int error = eden_login(eden, "deepthought");
    printf(" Attempting to send eden command \"test=2112;\" ...\n");
    eden_run(eden, "deepthought", "eden", "test=2112;");
    close(eden);
}

void test_eden_set_int() {
    /* Checking whether this actually works involves interrogating the
     * server side manually; and possibly watching the traffic with
     * a sniffer tool */
    printf("Ensure we can set an integer the easy way.\n");
    printf(" Connecting...\n");
    struct sockaddr_in ea = eden_sockaddr_in("192.168.1.6", EDEN_PORT);
    SOCKFD eden = connect_to_eden(ea);
    printf(" Logging in...\n");
    int error = eden_login(eden, "deepthought");
    printf(" Attempting to set foo to 15...\n");
    eden_set_int(eden, "deepthought", "foo", 15);
    close(eden);
}

int main() {
    test_eden_sockaddr_in();
    //test_server_conn();
    test_eden_cmd_wrap();
    //test_eden_login();
    //test_eden_run();
    test_eden_set_int();
    return 0;
}
