#include <stdio.h>
#include <ctype.h>
#include "ics.h"
#include "ashcurseshack.h"

/*  The following macros are for bit extraction from 16 bit words   */

#define func(A)  getbits(A,15,5)
#define mode(A)  getbits(A,10,2)
#define oper(A)  getbits(A,8,9)

/* Global variables and external variables used throughout the file */

extern int  SCR, V, Z, N, reg_x, reg_B, reg_A;
extern int  store(), save();
extern int  traceflag, execflag, maxorders, maxcount;
extern int  octcode[];
extern int  eflag;

/* exponential function - raises x to the power of n	 */

power(x, n)
    int         x, n;
{
    int         i, p;

    p = 1;
    for (i = 1; i <= n; ++i)
	p = p * x;
    return (p);
}

/*  Bit extraction procedure - obtains 'n' bits starting from position
    'p', from the 16 bit word 'x'.	 */

unsigned int
getbits(x, p, n)
    unsigned    x, p, n;
{
    return ((x >> (p + 1 - n)) & ~(~0 << n));
}


/* This procedure translates 'numb' from an unsigned value into
  a signed value and returns the new value.	*/

int
modif(numb)
    int         numb;
{
    int         current;

    if (numb >= 0)
	current = numb - 65536 * (numb / 65536);
    else
	current = numb - 65536 * (numb / 65536 - 1);
    if (current >= 32768)
	current = current - 65536;
    return (current);
}

/* This procedure translates 'numb' from a signed value into an unsigned
  value and retuns the new value.	  */

TV(numb)
    int         numb;
{
    int         newnumb;

    newnumb = numb;
    if (numb < 0)
	newnumb = numb + 65536;
    return (newnumb);
}


/* This procedure prints out the two characters represented by the
  value 'numb'. It only prints letters or digits	 */
canprint(numb)
    int         numb;
{
    char        fchar, lchar;

    fchar = numb / 256;
    lchar = numb % 256;
    if (isdigit(fchar) || isalpha(fchar))
      printw(" %c", fchar);
    else
      printw("  ");
    if (isdigit(lchar) || isalpha(lchar))
      printw("%c", lchar);
    refresh();
}

/* Prints out error messages if messtyp > 0.
Prints out other kinds of messages if messtyp < 0.  */

mess(messtyp)
    int         messtyp;
{
    if (messtyp > 7) {
	printw("\nEXECUTION OF ICS3 MACHINE FAILED AT LOCATION %d:", --SCR / PAG);
	printOs(8, SCR % PAG, 3, true);
    }
    if (messtyp > 0)
	printw("\n ERROR MESSAGE :-    ");
    switch (messtyp) {
    case 1:
	printw(" MNEMONIC UNRECOGNISED \n\n");
	break;
    case 2:
	printw(" MODE FIELD UNRECOGNISED \n\n");
	break;
    case 3:
	printw(" OCTAL CODE UNRECOGNISED \n\n");
	break;
    case 4:
	printw(" INVALID PAGE NUMBER \n\n");
	break;
    case 5:
	printw(" ORDER CODE INVALID \n\n");
	break;
    case 6:
	printw(" OUT OF MEMORY SPACE\n\n ");
	break;
    case 7:
	printw(" INTEGER VALUE INVALID\n\n");
	break;
    case 8:
	printw(" NONE INTEGER INPUT\n\n");
	printw("           >: ");
	break;
    case 10:
	printw(" ZERO ORDER FOUND\n\n ");
	break;
    case 11:
	printw(" ATTEMPT TO OPERATE ON BUS ADDRESS \n\n");
	break;
    case 12:
	printw(" INCORRECT PACKED CHARACTER FORMAT\n\n");
	printw("          >:");
	break;
    case 13:
	printw(" SCR OUT OF MEMORY SPACE\n\n ");
	break;
    case 14:
	printw(" exceeds max no of orders\n\n");
	break;
    case 15:
	printw(" ORDER CODE NOT RECOGNISED\n\n");
	break;
    case 16:
	printw(" MODE FIELD CAN ONLY TAKE VALUE '0'\n\n");
	break;
    case 17:
	printw(" MODE FIELD UNRECOGNISED\n\n");
	break;
    case 18:
	printw(" MODE FIELD CAN ONLY TAKE VALUES '1' - '3'\n\n");
	break;
    case -1:
	printw(" NORMAL EXIT\n\n");
	break;
    }
    refresh();
    eflag = 0;
}

/* Checks that an input order code is a valid code value	*/

int
isoctcode(valu)
    int         valu;
{
    int         isoct = 0, i;

    for (i = 0; i < 28; i++)
	if (valu == octcode[i])
	    isoct = true;
    return (isoct);
}

/* Obtains data requested during the execution of a program	*/

getdata()
{
    int         counter, symbl, stayin;

    do {
	stayin = false;
	symbl = getchar();
	while (symbl == '\t' || symbl == ' ')
	    symbl = getchar();
	switch (symbl) {
	case '+':
	    counter = 0;
	    symbl = getchar();
	    while (isdigit(symbl)) {
		counter = counter * 10 + (symbl - '0');
		symbl = getchar();
	    }
	    if (symbl != '\n')
		eflag = 8;
	    break;
	case '-':
	    counter = 0;
	    symbl = getchar();
	    while (isdigit(symbl)) {
		counter = counter * 10 + (symbl - '0');
		symbl = getchar();
	    }
	    if (symbl != '\n')
		eflag = 8;
	    counter = (-1 * counter) + 65536;
	    break;
	default:
	    eflag = 8;
	    break;
	}
	if (eflag == 8) {
	    stayin = true;
	    SCR++;
	}
	mess(eflag);
	while (symbl != '\n')
	    symbl = getchar();
    } while (stayin);
    return (counter);
}

/* Obtains packed character type data requested during the
execution of a program.  */

int
getchars()
{
    int         stayin = 0, packd = 0, keyin = 0;

    do {
	packd = 0;
	keyin = getchar();
	while (keyin == '\t' || keyin == ' ')
	    keyin = getchar();
	eflag = 12;
	if (keyin == '"')
	    if ((keyin = (getchar())) != '\n') {
		packd = 256 * keyin;
		if ((keyin = getchar()) != '\n') {
		    packd = packd + keyin;
		    if ((keyin = (getchar())) == '"')
			eflag = 0;
		}
	    }
	if (eflag == 12) {
	    SCR++;
	    mess(eflag);
	} else
	    stayin = 1;
	while (keyin != '\n')
	    keyin = getchar();
    } while (stayin == 0);
    return (packd);
}


/* Prints out a number(numb) in base(base) of length(space).
(all) determines whether it should print leading zeros */

printOs(base, numb, space, all)
    int         base, numb, space, all;
{
    int         k, q;

    if (numb < 0) {
	printw("-");
	numb = (-1) * numb;
    } else
	printw(" ");
    for (k = space - 1; k >= 0; k--) {
	q = numb / power(base, k);
	numb = numb % power(base, k);
	if (q != 0)
	    all = true;
	if (all == true)
	    printw("%d", q);
	else {
	    if (k == 0)
		printw("0");
	    else
		printw(" ");
	}
    }
    refresh();
}

/* translates a number(subject) from decimal into octal.'length'
indicates 'the number of digits - 1' in the number */

int
dec_oct(subject, length)
    int         subject, length;
{
    int         temp, numb, l, real = 0;

    numb = subject;
    for (l = length; l >= 0; l--) {
	temp = (numb / (power(8, l))) * (power(10, l));
	real = real + temp;
	numb = numb % (power(8, l));
    }
    return (real);
}


/* Translates a number(subject)  from octal into decimal.'length'
indicates the 'number of digits - l' in the number */

int
oct_dec(subject, length)
    int         subject, length;
{
    int         temp, numb, l, real = 0;

    numb = subject;
    for (l = length; l >= 0; l--) {
	temp = (numb / (power(10, l))) * (power(8, l));
	real = real + temp;
	numb = numb % (power(10, l));
    }
    return (real);
}


/* Changes an order format into one decimal 16 bit equivelent number. 'argl'
is the order code,'arg2' is the mode value and 'arg3' is the operand value. */

int
dec(argl, arg2, arg3)
    int         argl, arg2, arg3;
{
    int         s;

    s = argl;
    s = s * 4 + arg2;
    s = (s * 512) + arg3;
    return (s);
}

/* Logical AND ( Aorx = 10) or logical OR ( Aorx = 12) operation. 'opernd is
the value of the operand. 'regster' is the value in the general register
(reg_A) or the index register (reg_x). */

int
andor(opernd, regster, Aorx)
    int         opernd, regster, Aorx;
{
    int         tempo, tempr, count, reslt = 0;

    for (count = 0; count < 16; count++) {
	tempo = opernd / 2;
	tempr = regster / 2;
	switch (Aorx) {
	case 10:
	    if ((tempo * 2 == opernd) && (tempr * 2 == regster))
		reslt = reslt + power(2, count);
	    break;
	case 12:
	    if ((tempo * 2 == opernd) || (tempr * 2 == regster))
		reslt = reslt + power(2, count);
	    break;
	}
	opernd = tempo;
	regster = tempr;
    }
    return (reslt);
}


/* Logical NOT operation by bit inversion. 'regster' is the value in the general
register (reg_A) or the index register (reg_x). */

int
notA(regster)
    int         regster;
{
    int         tempr, count, reslt = 0;

    for (count = 1; count < 15; count++) {
	tempr = regster / 2;
	if (regster == tempr * 2)
	    reslt = reslt + power(2, count);
	regster = tempr;
    }
    return (reslt);
}


/* This sets the condition codes after the execution of an order.
'reg' is the value in the register after the execution of an
order. 'overfl' is at value '0' unless an order requires the
overflow flag 'V' to be set after it has been executed.  */

setcc(reg, overfl)
    int         reg, overfl;
{
    Z = N = V = 0;
    if (reg < 0)
	N = 1;
    if (reg == 0)
	Z = 1;
    if (overfl == true && (-32768 > reg || reg > 32767))
	V = 1;
}

/* This produces in table form the contents of the memory in the range
'first' to 'last'. It then prints out the contents of The buses and registers
followed by the condit ion codes and
the value of the SCR.
*/

dump(first, last)
    int         first, last;
{
    int         i, j, k, start, ends;
    int         sit;

    printw(" ICS STORE DUMP FROM	");
    printw("%d:", (first / 512));
    printOs(8, (first % 512), 3, true);
    printw(" TO %d:", (last / 512));

    printOs(8, (last % 512), 3, true);
    printw("\n");
    printw(" store locations containing zero are supressed \n");
    printw(" PAGE    LOCATION   DECIMAL    ");
    printw(" BINARY             ORDER FORMAT CHAR\n");
    start = first;
    ends = fini(512, last);
    for (j = (start % 1024) / 512; j < 2; j++) {
	printw("  %d:\n", j);
	for (i = start; i <= ends; i++) {
	    if ((store(i + (j * PAG)) != NULL) && (i < PAG + (j * PAG))) {
		sit = i + (j * PAG);
		printw("           ");
		printOs(8, (i % PAG), 3, true);
		printw("     ");
		printOs(10, modif(store(sit)), 5, false);
		printw("     ");
		printOs(2, store(sit), 16, true);
		printw("    ");
		printOs(8, func(store(sit)), 2, true);
		printw("    %d   ", mode(store(sit)));
		printOs(8, oper(store(sit)), 3, true);
		canprint(store(sit));
		printw("\n");
	    }
	}
	start = 0;
	ends = fini((last - 512), 512);
    }
    printw(" BUSES \n 63: \n        ");
    for (k = 0; k < 4; k++)
	if (store(1025 + k) != 0) {
	    printOs(10, (BUSCI + k), 5, false);
	    printw(" ");
	    printOs(10, modif(store(1025 + k)), 5, false);
	    printw(" ");
	    printOs(2, store(1025 + k), 16, true);
	}
    printw("\n REGISTERS \n        A ");
    printOs(10, modif(reg_A), 5, false);
    printw(" ");
    printOs(2, reg_A, 16, true);
    printw("\n        B ");
    printOs(10, modif(reg_B), 5, false);
    printw(" ");
    printOs(2, reg_B, 16, true);
    printw("\n        x ");
    printOs(10, modif(reg_x), 5, false);
    printw(" ");
    printOs(2, reg_x, 16, true);
    printw("\n\n        SCR = %d:", (SCR / PAG));
    printOs(8, (SCR % PAG), 3, true);
    printw("      V = %d    Z = %d    N = %d\n\n\n", V, Z, N);
    refresh();
}

/* This finds the effective value of the operand depending on first the
order code and secondly the mode field. 'pos' is the position in
memory at which the current order is stored. */

int
findeffect(pos)
    int         pos;
{
    int         effect = 1, temp;

    switch (func(store(pos))) {
    case 16:
    case 17:
	switch (mode(store(pos))) {
	case 0:
	    effect = oper(store(pos));
	    break;
	default:
	    eflag = 16;
	    break;
	}
	break;
    case 0:
    case 1:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8:
    case 9:
    case 10:
    case 11:
    case 12:
    case 13:
    case 14:
    case 15:
    case 20:
    case 21:
    case 23:
	switch (mode(store(pos))) {
	case 0:
	    effect = oper(store(pos));
	    break;
	case 1:
	    effect = store(oper(store(pos)));
	    break;
	case 2:
	    effect = store(oper(store(pos)) + reg_x);
	    break;
	case 3:
	    if (store(oper(store(pos))) == BUSDI) {
		execflag = 1;
		effect = 1027;
	    }
	    if (store(oper(store(pos))) == BUSCI) {
		execflag = 2;
		effect = 1025;
	    }
	    if (execflag == 0)
		if ((temp = store(oper(store(pos)))) < 1024)
		    effect = store(temp);
	    break;
	default:
	    eflag = 17;
	    break;
	}
	break;
    case 2:
    case 3:
    case 24:
    case 25:
    case 26:
    case 27:
    case 28:
    case 29:
    case 30:
	switch (mode(store(pos))) {
	case 1:
	    effect = oper(store(pos));
	    break;
	case 2:
	    effect = oper(store(pos)) + reg_x;
	    break;
	case 3:
	    if (store(oper(store(pos))) == BUSDO) {
		execflag = 1;
		effect = 1028;
	    }
	    if (store(oper(store(pos))) == BUSCO) {
		execflag = 2;
		effect = 1026;
	    }
	    if ((execflag == 0) && (oper(store(pos)) < 1024))
		effect = store(oper(store(pos)));
	    break;
	default:
	    eflag = 18;
	    break;
	}
	break;
    case 31:
	effect = 0;
	break;
    default:
	eflag = 15;
	break;
    }
    return (effect);
}

/* This provides the optional trace facility allowing the user to monitor the
operation of their program during execution.'tra' is the position in
memory at which the current order is stored. */

trace(tra)
    int         tra;
{
    int         effect;

    effect = findeffect(tra);
    switch (func(store(tra))) {
    case 16:
    case 17:
	printw("  LITERAL OPERAND = ");
	printOs(2, store(tra), 16, true);
	printw("\n\n");
	break;
    case 2:
    case 3:
    case 24:
    case 25:
    case 26:
    case 27:
    case 28:
    case 29:
    case 30:
    case 31:
	printw("  OPERAND = ");
	printOs(2, effect, 16, true);
	printw("\n\n");
	break;
    case 0:
    case 1:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8:
    case 9:
    case 10:
    case 11:
    case 12:
    case 13:
    case 14:
    case 15:
    case 20:
    case 21:
    case 23:
	printw(" M = %d A = ", mode(store(tra)));
	printOs(2, store(tra), 16, true);
	printw("\n ADDRESS = %d: ", tra / PAG);
	printOs(8, tra % PAG, 3, true);
	printw("\n");
	break;
    }
    refresh();
}


/* This procedure runs the program by executing all the orders in
the program code.  */

exec()
{
    int         effect, AB, temp, trSCR;

    Z = N = V = maxcount = 0;
    printf("\n ENTER EXEC ....\n ");
    do {
	trSCR = SCR;
/*
	if (isorder(SCR) == true) {
*/
	    execflag = 0;
	    effect = findeffect(SCR);
	    switch (func(store(SCR))) {
	    case 0:
		switch (execflag) {
		case 1:
		    printf("		   >: ");
		    reg_A = getdata();
		    save(effect, TV(reg_A));
		    break;
		case 0:
		    reg_A = effect;
		    break;
		case 2:
		    printf("          >: ");
		    reg_A = getchars();
		    break;
		}
		setcc(modif(reg_A), 0);
		break;
	    case 1:
		switch (execflag) {
		case 1:
		    printf("	>: ");
		    reg_x = getdata();
		    save(effect, TV(reg_A));
		    break;
		case 0:
		    reg_x = effect;
		    break;
		case 2:
		    printf("          >: ");
		    reg_x = getchars();
		    break;
		}
		setcc(modif(reg_x), 0);
		break;
	    case 2:
		switch (execflag) {
		case 1:
		    printf("\n");
		    printOs(10, modif(reg_A), 5, false);
		    printf("\n");
		    save(effect, TV(reg_A));
		    break;
		case 0:
		    save(effect, TV(reg_A));
		    /* saveorder(effect, 0); */
		    break;
		case 2:
		    canprint(reg_A);
		    break;
		}
		break;
	    case 3:
		switch (execflag) {
		case 1:
		    printf("\n");
		    printOs(10, modif(reg_x), 5, false);
		    printf("\n");
		    save(effect, reg_x);
		    break;
		case 0:
		    save(effect, TV(reg_x));
		    /* saveorder(effect, 0); */
		    break;
		case 2:
		    canprint(reg_x);
		    break;
		}
		break;
	    case 4:
		reg_A = TV(modif(reg_A) + modif(effect));
		setcc(reg_A, 1);
		break;
	    case 5:
		reg_x = TV(modif(reg_x) + modif(effect));
		setcc(modif(reg_x), 1);
		break;
	    case 6:
		reg_A = TV(modif(reg_A) - modif(effect));
		setcc(modif(reg_A), 1);
		break;
	    case 7:
		reg_x = TV(modif(reg_x) - modif(effect));
		setcc(modif(reg_x), 1);
		break;
	    case 8:
		setcc(modif(reg_A) - modif(effect), 0);
		break;
	    case 9:
		setcc(modif(reg_x) - modif(effect), 0);
		break;
	    case 10:
	    case 12:
		reg_A = andor(effect, reg_A, func(store(SCR)));
		setcc(modif(reg_A), 0);
		break;
	    case 11:
	    case 13:
		reg_x = andor(effect, reg_x, func(store(SCR)));
		setcc(modif(reg_A), 0);
		break;
	    case 14:
		reg_A = notA(reg_A);
		setcc(modif(reg_A), 0);
		break;
	    case 15:
		reg_x = notA(reg_x);
		setcc(modif(reg_x), 0);
		break;
	    case 16:
		if (effect > 32)
		    reg_A = reg_B = 0;
		else {
		    if (effect > 16)
			reg_A = 0;
		    reg_A = TV(modif(reg_A) * power(2, effect));
		    reg_B = TV((modif(reg_B) * power(2, effect)) +
			       (modif(reg_A) / power(2, 16)));
		    reg_A = TV(modif(reg_A) % power(2, 16));
		    reg_B = TV(modif(reg_B) % power(2, 16));
		}
		setcc(modif(reg_A), 0);
		break;
	    case 17:
		if (effect > 32)
		    reg_A = reg_B = 0;
		else {
		    temp = (modif(reg_B) * power(2, 16)) + modif(reg_A);
		    temp = TV(temp / power(2, effect));
		    reg_B = TV(temp / power(2, 16));
		    reg_A = TV(temp % power(2, 16));
		}
		setcc(modif(reg_A), 0);
		break;
	    case 20:
		reg_B = TV(((effect * modif(reg_A)) / power(2, 16)) % power(2, 16));
		reg_A = (effect * reg_A) % power(2, 16);
		break;
	    case 21:
		if (effect != 0) {
		    AB = (reg_B * power(2, 16)) + reg_A;
		    reg_B = (AB % effect) % power(2, 16);
		    reg_A = (AB / effect) % power(2, 16);
		    setcc(modif(reg_A), 1);
		} else
		    V = 1;
		break;
	    case 23:
		temp = reg_A;
		reg_A = reg_B;
		reg_B = temp;
		break;
	    case 24:
		if (Z == 0 && N == 0)
		    SCR = effect - 1;
		break;
	    case 25:
		if (N == 1)
		    SCR = effect - 1;
		break;
	    case 26:
		if (Z == 1)
		    SCR = effect - 1;
		break;
	    case 27:
		if (V == 1)
		    SCR = effect - 1;
		break;
	    case 28:
		SCR = effect - 1;
		break;
	    case 29:
		reg_x = SCR + 1;
		SCR = effect - 1;
		break;
	    case 30:
		if ((--reg_x) != 0)
		    SCR = effect - 1;
		break;
	    case 31:
		eflag = -1;
		break;
	    default:
		eflag = 15;
		break;
	    }
	    if (traceflag == true)
		trace(trSCR);
/*
	}
*/
	SCR++;
	if ((store(SCR) == 0) && (store(SCR - 1) == 0))
	    eflag = 10;
	if ((++maxcount) > maxorders)
	    eflag = MAXORDER;
	if (SCR > 1024)
	    eflag = 13;
	if ((SCR < 1029) && (SCR > 1024))
	    eflag = 11;
    } while (eflag == 0);
    mess(eflag);
}
